#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
seqeunce 转换成 vector

Authors: PanXu
Date:    2020/07/08 20:09:00
"""

from .cnn_seq2vec import CnnSeq2Vec
from .attention_seq2vec import AttentionSeq2Vec
from .rnn_seq2vec import RnnSeq2Vec
